#This is the class for construcing FSAs.  The individual CAs (both inputs
#and states) run at 5 ms.  There may be problems if they run faster or slower.
#The class enables you to build 2 input FSAs.  You can't go from state A
#to state B, then directly back.
#There are test functions at the end. 

import nealParams as nealParameters

if (nealParameters.pynnVersion=="pynn")and(nealParameters.simulator=="spinnaker"):
    import pyNN.spiNNaker as sim
elif (nealParameters.pynnVersion=="pynn")and(nealParameters.simulator=="nest"):
    import pyNN.nest as sim
    import cPickle as pickle
elif (nealParameters.pynnVersion=="pynest")and(nealParameters.simulator=="nest"):
    import nest as nest

import numpy as np

from nealCoverClass import NealCoverFunctions

class FSAHelperFunctions:
    def __init__(self, simName):
        self.simulator = simName
        self.neal = NealCoverFunctions(self.simulator)
        self.initParams()

    #FSA Parmaeters
    def initParams(self):
        self.CA_SIZE = 5  #This will (almost certainly) not work with a 
                     #different sized CA.

        if (nealParameters.pynnVersion == "pynn"):
            self.INPUT_WEIGHT = 0.1
            if (nealParameters.simulator == 'nest'):
                self.INTRA_CA_WEIGHT = 0.025
            if (nealParameters.simulator == 'spinnaker'):
                self.INTRA_CA_WEIGHT = 0.03 
            self.HALF_ON_WEIGHT = 0.0017 
            self.FULL_ON_WEIGHT = 0.01
            self.CA_STOPS_CA_WEIGHT = -0.3 
            self.CELL_PARAMS = {'v_thresh':-55.0, 'v_reset' : -70.0, 
                                'tau_refrac': 2.0 , 'tau_syn_E': 5.0,  
                                #'e_rev_I': -100.0, 
                                #'tau_syn_I': 1.0, #bug2 fix
                                'v_rest' : -65.0,'i_offset':0.0}

        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            self.INPUT_WEIGHT = 36.0
            self.INTRA_CA_WEIGHT = 8.0
            #self.HALF_ON_WEIGHT = 2.0 
            self.HALF_ON_WEIGHT = 1.8 
            self.FULL_ON_WEIGHT = 4.0
            self.CA_STOPS_CA_WEIGHT = -20.0 
            self.CELL_PARAMS = {'V_reset': -70.0, 't_ref':2.0, 'tau_syn_ex':5.0,
                                'g_L':40.0}


    #--------Finite State Automata Functions ------------
    #---Functions that turn on states by one item

    #-- Function to ignite a state from a spike source
    #-- Uses INPUT_WEIGHT
    #use this when your using a spikesource
    def turnOnStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        connector = []
        for toOffset in range (0,self.CA_SIZE):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,self.INPUT_WEIGHT,
                                      nealParameters.DELAY)]
        self.neal.nealProjection(spikeSource, toNeurons, connector,'excitatory')

    def turnOnStateFromOneNeuron(self,fromNeurons,fromNeuron, toNeurons, toCA):
        connector = []
        for toOffset in range (0,self.CA_SIZE):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,self.INPUT_WEIGHT,
                                      nealParameters.DELAY)]
        self.neal.nealProjection(fromNeurons, toNeurons, connector,'excitatory')

    #Function to turn on one state from another
    #Call with fromPopulation, fromCA, toPopulation and toCA
    def turnOnStateFromState(self,fromNeurons,fromCA,toNeurons,toCA):
        connector = []
        #uses FULL_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.FULL_ON_WEIGHT,
                                          nealParameters.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')


    #---Create a CA that will persistently fire.
    #-- Assumes neurons in the same population
    #-- Uses INTRA_CA_WEIGHT
    def makeCA(self,neurons, CA):
        #print 'makeCA
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (CA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (CA*self.CA_SIZE)
                if (toNeuron != fromNeuron):
                    connector = connector + [(fromNeuron,toNeuron,
                                              self.INTRA_CA_WEIGHT,
                                              nealParameters.DELAY)]

        self.neal.nealProjection(neurons,neurons,connector,'excitatory')

    #--Functions when two inputs are needed to turn on a third
    #-- Two states are needed to turn on a third.  
    #-- This connects one of the inputs to the the third.
    #-- Uses HALF_ON_WEIGHT
    def stateHalfTurnsOnState(self,fromNeurons,fromCA,toNeurons,toCA):
        connector = []
        #uses HALF_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.HALF_ON_WEIGHT,
                                          nealParameters.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    #---- One State or other set of neurons turns off another
    #-- Uses CA_STOPS_CA_WEIGHT
    def stateTurnsOffState(self,fromNeurons, fromCA, toNeurons, toCA):
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,
                                              self.CA_STOPS_CA_WEIGHT,
                                              nealParameters.DELAY)]

        self.neal.nealProjection(fromNeurons,toNeurons,connector,'inhibitory')


#------test functions
    #initialize the simulator. 
    def testInit(self):
        if (nealParameters.pynnVersion == "pynn"):
            #print "spin" or nest
            sim.setup(timestep=nealParameters.DELAY,
                    min_delay=nealParameters.DELAY,
                    max_delay=nealParameters.DELAY, debug=0)

        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            nest.ResetKernel()
            nest.SetKernelStatus({'resolution':nealParameters.DELAY})


    def testCreateTwoInputs(self):
        inputSpikeTimes0 = [10.0]
        inputSpikeTimes1 = [50.0]
        if (nealParameters.pynnVersion == "pynn"):
            spikeArray0 = {'spike_times': [inputSpikeTimes0]}
            spikeGen0=sim.Population(1,sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')
            spikeArray1 = {'spike_times': [inputSpikeTimes1]}
            spikeGen1=sim.Population(1, sim.SpikeSourceArray, spikeArray1,
                                   label='inputSpikes_1')

        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            spikeGen0 = nest.Create('spike_generator',
                #params = {'spike_times':np.array([10.0,15.0, 20.0,25.0,30.0])})
                params = {'spike_times': np.array(inputSpikeTimes0)})

            spikeGen1 = nest.Create('spike_generator',
                params = {'spike_times': np.array(inputSpikeTimes1)})

            spikeDet = nest.Create('spike_detector')

        else: print "bad simulator for spike generator"

        return [spikeGen0,spikeGen1]

    def testCreateNeurons(self):
        if (nealParameters.pynnVersion == "pynn"): 
            if (nealParameters.simulator == 'nest'):
                numNeurons = 15
            elif (nealParameters.simulator == 'spinnaker'):
                numNeurons = 100
        
            cells=sim.Population(numNeurons,sim.IF_cond_exp,self.CELL_PARAMS)
        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            #in python Models() to see all the models including these
            #print GetDefaults (neuronType)

            cells = nest.Create("iaf_cond_exp",n=15,params = self.CELL_PARAMS)

        return cells

    def testCreateRecorder(self):
        spikeDet = 0
        if (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            spikeDet = nest.Create('spike_detector')

        #with spinnaker you can just set record on the cells
        return spikeDet

    def testSetupRecording(self,cells, spikeDetector):
        if (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'nest'):
            cells.record({'spikes','v'})
        elif (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'spinnaker'):
            cells.record()
        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            nest.Connect(cells,spikeDetector)

    def test3StateFSA(self, firstSpikeGenerator, secondSpikeGenerator,
                      stateCells):
        #Build the FSA
        self.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
        self.turnOnStateFromSpikeSource(secondSpikeGenerator,stateCells,1)
        self.makeCA(stateCells,0)
        self.makeCA(stateCells,1)
        self.makeCA(stateCells,2)
        self.stateHalfTurnsOnState(stateCells,0,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,0)
        #comment below out to check state 0 alone does not turn on state 2
        self.stateHalfTurnsOnState(stateCells,1,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,1)

    def testRunFSA(self,duration):
        if (nealParameters.pynnVersion == "pynn"):
            sim.run(duration)
        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            nest.Simulate(duration)


    def testPrintPklSpikes(self,fileName):
        fileHandle = open(fileName)
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    print neuronNum, spikes[spike]
        fileHandle.close()
    

    def testPrintResults(self,simCells,nestRecorder):
        #print
        if (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'nest'):
            simCells.printSpikes('temp.pkl')
            self.testPrintPklSpikes('temp.pkl')
        elif (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'spinnaker'):
            simCells.printSpikes('temp.sp')
        elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
            spikes = nest.GetStatus(nestRecorder)
            testTimes = spikes[0]['events']['times']
            testSenders = spikes[0]['events']['senders']
            print 'spikes'
            for spikeNumber in range (0,len(testTimes)):
                print testTimes[spikeNumber], testSenders[spikeNumber]
